"""Load AWS IP address ranges."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import hashlib
import json
import urllib.request
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

from awsipranges.config import (
    AWS_IP_ADDRESS_RANGES_URL,
    CREATE_DATE_FORMAT,
    CREATE_DATE_TIMEZONE,
)
from awsipranges.exceptions import raise_for_status
from awsipranges.models.awsipprefix import aws_ip_prefix
from awsipranges.models.awsipprefixes import AWSIPPrefixes
from awsipranges.utils import check_type


def get_json_data(
    cafile: Union[str, Path, None] = None, capath: Union[str, Path, None] = None
) -> Tuple[Dict[str, Any], Optional[str]]:
    """Retrieve and parse the AWS IP address ranges JSON file."""
    check_type("cafile", cafile, (str, Path), optional=True)
    cafile = Path(cafile) if isinstance(cafile, str) else cafile
    if cafile and not cafile.is_file():
        raise ValueError(
            "Invalid path; cafile must be a path to a bundled CA certificate file."
        )

    check_type("capath", capath, (str, Path), optional=True)
    capath = Path(capath) if isinstance(capath, str) else capath
    if capath and not capath.is_dir():
        raise ValueError("Invalid path; capath must be a path to a directory.")

    with urllib.request.urlopen(
        AWS_IP_ADDRESS_RANGES_URL, cafile=cafile, capath=capath
    ) as response:
        raise_for_status(response)

        response_bytes = response.read()
        response_data = json.loads(response_bytes)

        if "md5" in hashlib.algorithms_available:
            md5_hash = hashlib.md5()
            md5_hash.update(response_bytes)
            md5_hex_digest = md5_hash.hexdigest()
        else:
            md5_hex_digest = None

    return response_data, md5_hex_digest


def get_ranges(cafile: Path = None, capath: Path = None) -> AWSIPPrefixes:
    """Get the AWS IP address ranges from the published JSON document.

    It is your responsibility to verify the TLS certificate presented by the
    server. By default, the Python
    [urllib](https://docs.python.org/3/library/urllib.html) module (used by
    this function) verifies the TLS certificate presented by the server against
    the system-provided certificate datastore.

    You can download the Amazon root CA certificates from the
    [Amazon Trust Services](https://www.amazontrust.com/repository/)
    repository.

    The optional `cafile` and `capath` parameters may be used to specify a set
    of trusted CA certificates for the HTTPS request. `cafile` should point to a
    single file containing a bundle of CA certificates, whereas `capath`
    should point to a directory of certificate files with OpenSSL hash filenames.
    To verify the TLS certificate against Amazon root certificates, download the
    CA certificates (in PEM format) from Amazon Trust Services and provide the
    path to the certificate(s) using the `cafile` or `capath` parameters.

    See the OpenSSL [SSL_CTX_load_verify_locations](https://www.openssl.org/docs/man1.1.0/man3/SSL_CTX_set_default_verify_dir.html)
    documentation for details on the expected CAfile and CApath file formats.

    **Parameters:**

    - **cafile** (_optional_ Path) - path to a file of stacked (concatenated) CA
      certificates in PEM format

    - **capath** (_optional_ Path) - path to a directory containing one or more
      certificates in PEM format using the OpenSSL subject-name-hash filenames

    **Returns:**

    The AWS IP address ranges in a `AWSIPPrefixes` collection.
    """
    json_data, json_md5 = get_json_data(cafile=cafile, capath=capath)

    assert "syncToken" in json_data
    assert "createDate" in json_data
    assert "prefixes" in json_data
    assert "ipv6_prefixes" in json_data

    return AWSIPPrefixes(
        sync_token=json_data["syncToken"],
        create_date=datetime.strptime(
            json_data["createDate"], CREATE_DATE_FORMAT
        ).replace(tzinfo=CREATE_DATE_TIMEZONE),
        ipv4_prefixes=(aws_ip_prefix(record) for record in json_data["prefixes"]),
        ipv6_prefixes=(aws_ip_prefix(record) for record in json_data["ipv6_prefixes"]),
        md5=json_md5,
    )



"""Work with AWS IP address ranges in native Python."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from awsipranges.data_loading import get_ranges  # noqa: F401
from awsipranges.exceptions import AWSIPRangesException, HTTPError  # noqa: F401
from awsipranges.models.awsipprefix import (  # noqa: F401
    AWSIPPrefix,
    AWSIPv4Prefix,
    AWSIPv6Prefix,
)
from awsipranges.models.awsipprefixes import AWSIPPrefixes  # noqa: F401



"""Package configurations."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import timezone


AWS_IP_ADDRESS_RANGES_URL = "https://ip-ranges.amazonaws.com/ip-ranges.json"

CREATE_DATE_FORMAT = "%Y-%m-%d-%H-%M-%S"
CREATE_DATE_TIMEZONE = timezone.utc



"""Model an AWS IP Prefix as a native Python object."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import itertools
from abc import ABC, abstractmethod
from functools import total_ordering
from ipaddress import (
    ip_network,
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from typing import Any, Dict, Iterable, Tuple, Union

from awsipranges.utils import check_type


@total_ordering
class AWSIPPrefix(ABC):
    """AWS IP Prefix."""

    __slots__ = ["_prefix", "_region", "_network_border_group", "_services"]

    _prefix: Union[IPv4Network, IPv6Network]
    _region: str
    _network_border_group: str
    _services: Tuple[str, ...]

    def __init__(
        self,
        prefix: Union[str, IPv4Network, IPv6Network],
        region: str,
        network_border_group: str,
        services: Union[str, Iterable[str]],
    ) -> None:
        super().__init__()

        check_type("prefix", prefix, (str, IPv4Network, IPv6Network))
        check_type("region", region, str)
        check_type("network_border_group", network_border_group, str)
        check_type("services", services, (str, tuple))

        self._prefix = ip_network(prefix) if isinstance(prefix, str) else prefix
        self._region = region
        self._network_border_group = network_border_group
        self._services = (
            (services,) if isinstance(services, str) else tuple(sorted(services))
        )

    @property
    @abstractmethod
    def prefix(self) -> Union[IPv4Network, IPv6Network]:
        """The public IP network prefix."""
        return self._prefix

    @property
    def region(self) -> str:
        """The AWS Region or `GLOBAL` for edge locations.

        The `CLOUDFRONT` and `ROUTE53` ranges are GLOBAL.
        """
        return self._region

    @property
    def network_border_group(self) -> str:
        """The name of the network border group.

        A network border group is a unique set of Availability Zones or Local
        Zones from where AWS advertises IP addresses.
        """
        return self._network_border_group

    @property
    def services(self) -> Tuple[str, ...]:
        """Services that use IP addresses in this IP prefix.

        The addresses listed for `API_GATEWAY` are egress only.

        The service `"AMAZON"` is not a service but rather an identifier used
        to get all IP address ranges - meaning that every prefix is contained in
        the subset of prefixes tagged with the `"AMAZON"` service. Some IP
        address ranges are only tagged with the `"AMAZON"` service.
        """
        return self._services

    def __repr__(self) -> str:
        """An executable string representation of this object."""
        return (
            f"{self.__class__.__name__}("
            f"{str(self._prefix)!r}, "
            f"region={self._region!r}, "
            f"network_border_group={self._network_border_group!r}, "
            f"services={self._services!r}"
            f")"
        )

    def __str__(self) -> str:
        """The IP prefix in CIDR notation."""
        return self._prefix.with_prefixlen

    @property
    def __tuple(self) -> tuple:
        """A tuple representation of the AWS IP prefix."""
        return self._prefix, self._region, self._network_border_group, self._services

    def __eq__(self, other) -> bool:
        """Compare for equality.

        This method allows comparisons between AWSIPPrefix objects, None,
        IPv4Network and IPv6Network objects, and strings that can be converted
        to IPv4Network and IPv6Network objects.

        **Raises:**

        A `ValueError` exception if the `other` object is a string and
        cannot be converted to an IPv4Network or IPv6Network object.

        A `TypeError` if the `other` object is of an unsupported type.
        """
        if other is None:
            return False

        if isinstance(other, str):
            other = ip_network(other)

        if isinstance(other, (IPv4Network, IPv6Network)):
            return self.prefix == other

        if isinstance(other, AWSIPPrefix):
            return self.__tuple == other.__tuple

        raise TypeError(
            f"Cannot compare an AWSIPPrefix object with an object of type"
            f" {other!r}."
        )

    def __hash__(self) -> int:
        return hash(self.__tuple)

    def __lt__(self, other) -> bool:
        """Comparison operator to facilitate sorting.

        **Sort order:**

        - IPv4 prefixes before IPv6 prefixes
        - IP network addresses in ascending order
        - IP prefix length in ascending order
        - Region in ascending order
        - Network border group in ascending order
        - Services in ascending order
        """
        # Allow comparison between AWSIPPrefixes and Python native IPv4 and IPv6
        # network objects.
        if isinstance(other, (IPv4Network, IPv6Network)):
            if self.prefix != other:
                return self.prefix < other
            else:
                return False

        # Compare two AWSIPPrefix objects
        if self.prefix.version != other.prefix.version:
            return self.prefix.version < other.prefix.version

        if self.prefix.network_address != other.prefix.network_address:
            return self.prefix.network_address < other.prefix.network_address

        if self.prefix.prefixlen != other.prefix.prefixlen:
            return self.prefix.prefixlen < other.prefix.prefixlen

        if self.region != other.region:
            return self.region < other.region

        if self.network_border_group != other.network_border_group:
            return self.network_border_group < other.network_border_group

        if self.services != other.services:
            return self.services < other.services

        return False

    def __contains__(
        self,
        item: Union[
            str,
            IPv4Address,
            IPv4Network,
            IPv4Interface,
            IPv6Address,
            IPv6Network,
            IPv6Interface,
        ],
    ) -> bool:
        check_type(
            "item",
            item,
            (
                str,
                IPv4Address,
                IPv4Network,
                IPv4Interface,
                IPv6Address,
                IPv6Network,
                IPv6Interface,
            ),
        )

        item_network = ip_network(item, strict=False)

        if item_network.version != self.version:
            return False

        return item_network.subnet_of(self._prefix)

    @property
    def version(self) -> int:
        """The IP version (4, 6)."""
        return self._prefix.version

    @property
    @abstractmethod
    def network_address(self) -> Union[IPv4Address, IPv6Address]:
        """The network address for the network."""
        return self._prefix.network_address

    @property
    def prefixlen(self) -> int:
        """Length of the network prefix, in bits."""
        return self._prefix.prefixlen

    @property
    def with_prefixlen(self) -> str:
        """A string representation of the IP prefix, in network/prefix notation."""
        return self._prefix.with_prefixlen

    @property
    @abstractmethod
    def netmask(self) -> Union[IPv4Address, IPv6Address]:
        """The net mask, as an IP Address object."""
        return self._prefix.netmask

    @property
    def with_netmask(self) -> str:
        """A string representation of the network, with the mask in net mask notation."""
        return self._prefix.with_netmask

    @property
    @abstractmethod
    def hostmask(self) -> Union[IPv4Address, IPv6Address]:
        """The host mask (aka. wildcard mask), as an IP Address object."""
        return self._prefix.hostmask

    @property
    def with_hostmask(self) -> str:
        """A string representation of the network, with the mask in host mask notation."""
        return self._prefix.with_hostmask

    @property
    def num_addresses(self) -> int:
        """The total number of addresses in the network."""
        return self._prefix.num_addresses

    def __getattr__(self, name: str) -> Any:
        """Proxy all other attributes to self._prefix."""
        return getattr(self._prefix, name)


class AWSIPv4Prefix(AWSIPPrefix):
    """AWS IPv4 Prefix."""

    __slots__ = ["_prefix", "_region", "_network_border_group", "_services"]

    _prefix: IPv4Network

    def __init__(
        self,
        prefix: Union[str, IPv4Network],
        region: str,
        network_border_group: str,
        services: Union[str, Iterable[str]],
    ) -> None:
        super().__init__(
            prefix=prefix,
            region=region,
            network_border_group=network_border_group,
            services=services,
        )
        check_type("prefix", self._prefix, IPv4Network)

    @property
    def prefix(self) -> IPv4Network:
        """The public IPv4 network prefix."""
        return self._prefix

    @property
    def ip_prefix(self) -> IPv4Network:
        """The public IPv4 network prefix.

        This is a convenience attribute to maintain API compatibility with the
        JSON attribute names.
        """
        return self._prefix

    @property
    def network_address(self) -> IPv4Address:
        """The network address for the network."""
        return self._prefix.network_address

    @property
    def netmask(self) -> IPv4Address:
        """The net mask, as an IPv4Address object."""
        return self._prefix.netmask

    @property
    def hostmask(self) -> IPv4Address:
        """The host mask (aka. wildcard mask), as an IPv4Address object."""
        return self._prefix.hostmask


class AWSIPv6Prefix(AWSIPPrefix):
    """AWS IPv6 Prefix."""

    __slots__ = ["_prefix", "_region", "_network_border_group", "_services"]

    _prefix: IPv6Network

    def __init__(
        self,
        prefix: Union[str, IPv6Network],
        region: str,
        network_border_group: str,
        services: Union[str, Iterable[str]],
    ) -> None:
        super().__init__(
            prefix=prefix,
            region=region,
            network_border_group=network_border_group,
            services=services,
        )
        check_type("prefix", self._prefix, IPv6Network)

    @property
    def prefix(self) -> IPv6Network:
        """The public IPv6 network prefix."""
        return self._prefix

    @property
    def ipv6_prefix(self) -> IPv6Network:
        """The public IPv6 network prefix.

        This is a convenience attribute to maintain API compatibility with the
        JSON attribute names.
        """
        return self._prefix

    @property
    def network_address(self) -> IPv6Address:
        """The network address for the network."""
        return self._prefix.network_address

    @property
    def netmask(self) -> IPv6Address:
        """The net mask, as an IPv6Address object."""
        return self._prefix.netmask

    @property
    def hostmask(self) -> IPv6Address:
        """The host mask (aka. wildcard mask), as an IPv6Address object."""
        return self._prefix.hostmask


def aws_ip_prefix(json_data: Dict[str, str]) -> Union[AWSIPv4Prefix, AWSIPv6Prefix]:
    """Factory function to create AWS IP Prefix objects from JSON data."""
    check_type("data", json_data, dict)
    assert "ip_prefix" in json_data or "ipv6_prefix" in json_data
    assert "region" in json_data
    assert "network_border_group" in json_data
    assert "service" in json_data

    if "ip_prefix" in json_data:
        return AWSIPv4Prefix(
            prefix=json_data["ip_prefix"],
            region=json_data["region"],
            network_border_group=json_data["network_border_group"],
            services=json_data["service"],
        )

    if "ipv6_prefix" in json_data:
        return AWSIPv6Prefix(
            prefix=json_data["ipv6_prefix"],
            region=json_data["region"],
            network_border_group=json_data["network_border_group"],
            services=json_data["service"],
        )


def combine_prefixes(
    prefixes: Iterable[Union[AWSIPv4Prefix, AWSIPv6Prefix]]
) -> Union[AWSIPv4Prefix, AWSIPv6Prefix]:
    """Combine multiple AWS IP prefix records into a single AWS IP prefix.

    The prefix records should have identical prefixes, regions, and network
    border groups. Only the services should be combined.
    """
    prefixes = list(prefixes)
    assert len(prefixes) > 1

    first = prefixes[0]

    check_type("prefixes", first, (AWSIPv4Prefix, AWSIPv6Prefix))
    for prefix in prefixes[1:]:
        check_type("prefixes", prefix, type(first))

    # Ensure prefixes are the same (only services are different) before combining
    if not all(
        [
            (prefix.prefix, prefix.region, prefix.network_border_group)
            == (first.prefix, first.region, first.network_border_group)
            for prefix in prefixes[1:]
        ]
    ):
        raise ValueError(
            "Cannot combine prefixes with different prefix, region, or "
            "network_border_group values."
        )

    # Combine the services
    services = set(
        itertools.chain.from_iterable((prefix.services for prefix in prefixes))
    )
    services = tuple(sorted(services))

    return type(first)(
        prefix=first.prefix,
        region=first.region,
        network_border_group=first.network_border_group,
        services=services,
    )



"""Model a set of AWS IP address prefixes as a python collection."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import itertools
import pprint
from bisect import bisect_left
from collections import defaultdict
from datetime import datetime
from ipaddress import (
    ip_network,
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from typing import FrozenSet, Iterable, Optional, Tuple, Union

from awsipranges.models.awsipprefix import (
    AWSIPv4Prefix,
    AWSIPv6Prefix,
    combine_prefixes,
)
from awsipranges.utils import check_type, normalize_to_set, supernets, validate_values


# Main class
class AWSIPPrefixes(object):
    """A collection of AWS IP address prefixes."""

    _sync_token: Optional[str]
    _create_date: Optional[datetime]
    _ipv4_prefixes: Tuple[AWSIPv4Prefix, ...]
    _ipv6_prefixes: Tuple[AWSIPv6Prefix, ...]
    _md5: Optional[str]

    _regions: Optional[FrozenSet[str]] = None
    _network_border_groups: Optional[FrozenSet[str]] = None
    _services: Optional[FrozenSet[str]] = None

    def __init__(
        self,
        sync_token: Optional[str] = None,
        create_date: Optional[datetime] = None,
        ipv4_prefixes: Iterable[AWSIPv4Prefix] = None,
        ipv6_prefixes: Iterable[AWSIPv6Prefix] = None,
        md5: Optional[str] = None,
    ) -> None:
        super().__init__()

        check_type("sync_token", sync_token, str, optional=True)
        check_type("create_date", create_date, datetime, optional=True)
        check_type("ipv4_prefixes", ipv4_prefixes, Iterable)
        check_type("ipv6_prefixes", ipv6_prefixes, Iterable)
        check_type("md5", md5, str, optional=True)

        self._sync_token = sync_token
        self._create_date = create_date
        self._ipv4_prefixes = self._process_prefixes(ipv4_prefixes)
        self._ipv6_prefixes = self._process_prefixes(ipv6_prefixes)
        self._md5 = md5

    @staticmethod
    def _process_prefixes(
        prefixes: Iterable[Union[AWSIPv4Prefix, AWSIPv6Prefix]],
    ) -> Tuple[Union[AWSIPv4Prefix, AWSIPv6Prefix], ...]:
        """Create a deduplicated sorted tuple of AWS IP prefixes."""
        collect_duplicates = defaultdict(list)
        for prefix in prefixes:
            collect_duplicates[prefix.prefix].append(prefix)

        deduplicated_prefixes = list()
        for prefixes in collect_duplicates.values():
            if len(prefixes) == 1:
                prefix = prefixes[0]
            else:
                prefix = combine_prefixes(prefixes)
            deduplicated_prefixes.append(prefix)

        deduplicated_prefixes.sort()

        return tuple(deduplicated_prefixes)

    def _get_prefix(
        self, prefix: Union[str, IPv4Network, IPv6Network]
    ) -> Union[None, AWSIPv4Prefix, AWSIPv6Prefix]:
        """Retrieve a specific prefix from the AWS IP address ranges."""
        check_type("prefix", prefix, (str, IPv4Network, IPv6Network))
        if isinstance(prefix, str):
            prefix = ip_network(prefix)

        if isinstance(prefix, IPv4Network):
            prefixes_collection = self.ipv4_prefixes
        elif isinstance(prefix, IPv6Network):
            prefixes_collection = self.ipv6_prefixes
        else:
            raise TypeError("`prefix` must be an IPv4Network or IPv6Network object.")

        # Retrieve the prefix from the collection
        index = bisect_left(prefixes_collection, prefix)
        if (
            index != len(prefixes_collection)
            and prefixes_collection[index].prefix == prefix
        ):
            return prefixes_collection[index]
        else:
            # Not found
            return None

    @property
    def sync_token(self) -> str:
        """The publication time, in Unix epoch time format."""
        return self._sync_token

    @property
    def syncToken(self) -> str:  # noqa
        """The publication time, in Unix epoch time format.

        This is a convenience attribute to maintain API compatibility with the
        JSON attribute names.
        """
        return self._sync_token

    @property
    def create_date(self) -> datetime:
        """The publication date and time, in UTC."""
        return self._create_date

    @property
    def createDate(self) -> datetime:  # noqa
        """The publication date and time, in UTC.

        This is a convenience attribute to maintain API compatibility with the
        JSON attribute names.
        """
        return self._create_date

    @property
    def ipv4_prefixes(self) -> Tuple[AWSIPv4Prefix, ...]:
        """The IPv4 prefixes in the collection."""
        return self._ipv4_prefixes

    @property
    def prefixes(self) -> Tuple[AWSIPv4Prefix, ...]:
        """The IPv4 prefixes in the collection.

        This is a convenience attribute to maintain API compatibility with the
        JSON attribute names.
        """
        return self._ipv4_prefixes

    @property
    def ipv6_prefixes(self) -> Tuple[AWSIPv6Prefix, ...]:
        """The IPv6 prefixes in the collection."""
        return self._ipv6_prefixes

    @property
    def md5(self) -> Optional[str]:
        """The MD5 cryptographic hash value of the ip-ranges.json file.

        You can use this value to verify the integrity of the downloaded file.
        """
        return self._md5

    def __repr__(self) -> str:
        return pprint.pformat(
            {
                "sync_token": self.sync_token,
                "create_date": self.create_date,
                "ipv4_prefixes": self.ipv4_prefixes,
                "ipv6_prefixes": self.ipv6_prefixes,
                "md5": self.md5,
            }
        )

    def __contains__(
        self,
        item: Union[
            str,
            IPv4Address,
            IPv6Address,
            IPv4Interface,
            IPv6Interface,
            IPv4Network,
            IPv6Network,
            AWSIPv4Prefix,
            AWSIPv6Prefix,
        ],
    ) -> bool:
        """Is the IP address, interface, or network in the AWS IP Address ranges?"""
        try:
            self[item]
        except KeyError:
            return False
        else:
            return True

    def __getitem__(
        self,
        item: Union[
            str,
            IPv4Address,
            IPv6Address,
            IPv4Network,
            IPv6Network,
            IPv4Interface,
            IPv6Interface,
            AWSIPv4Prefix,
            AWSIPv6Prefix,
        ],
    ):
        """Get the AWS IP address prefix that contains the IPv4 or IPv6 item.

        Returns the longest-match prefix that contains the provided item.

        **Parameters:**

        - **item** (str, IPv4Address, IPv6Address, IPv4Network, IPv6Network,
          IPv4Interface, IPv6Interface, AWSIPv4Prefix, AWSIPv6Prefix) - the item
          to retrieve from the collection

        **Returns:**

        The `AWSIPv4Prefix` or `AWSIPv6Prefix` that contains the provided `item`.

        **Raises:**

        A `KeyError` exception if the provided `item` is not contained in the
        collection.
        """
        if isinstance(item, (AWSIPv4Prefix, AWSIPv6Prefix)):
            network = item.prefix
        else:
            network = ip_network(item, strict=False)

        for supernet in supernets(network):
            supernet_prefix = self._get_prefix(supernet)
            if supernet_prefix:
                return supernet_prefix

        raise KeyError(
            f"{item!r} is not contained in this AWSIPAddressRanges collection."
        )

    def get(
        self,
        key: Union[
            str,
            IPv4Address,
            IPv6Address,
            IPv4Network,
            IPv6Network,
            IPv4Interface,
            IPv6Interface,
            AWSIPv4Prefix,
            AWSIPv6Prefix,
        ],
        default=None,
    ) -> Union[AWSIPv4Prefix, AWSIPv6Prefix]:
        """Get the AWS IP address prefix that contains the IPv4 or IPv6 key.

        Returns the longest-match prefix that contains the provided key or the
        value of the `default=` parameter if the key is not found in the
        collection.

        **Parameters:**

        - **key** (str, IPv4Address, IPv6Address, IPv4Network, IPv6Network,
          IPv4Interface, IPv6Interface, AWSIPv4Prefix, AWSIPv6Prefix) - the IP
          address or network to retrieve from the collection
        - **default** - the value to return if the key is not found in the
          collection

        **Returns:**

        The `AWSIPv4Prefix` or `AWSIPv6Prefix` that contains the provided key.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def get_prefix_and_supernets(
        self,
        key: Union[
            str,
            IPv4Address,
            IPv6Address,
            IPv4Network,
            IPv6Network,
            IPv4Interface,
            IPv6Interface,
            AWSIPv4Prefix,
            AWSIPv6Prefix,
        ],
        default=None,
    ) -> Optional[Tuple[Union[AWSIPv4Prefix, AWSIPv6Prefix], ...]]:
        """Get the prefix and supernets that contain the IPv4 or IPv6 key.

        Returns a tuple that contains the longest-match prefix and supernets
        that contains the provided key or the value of the `default=` parameter
        if the key is not found in the collection.

        The tuple is sorted by prefix length in ascending order (shorter prefixes
        come before longer prefixes).

        **Parameters:**

        - **key** (str, IPv4Address, IPv6Address, IPv4Network, IPv6Network,
          IPv4Interface, IPv6Interface, AWSIPv4Prefix, AWSIPv6Prefix) - the IP
          address or network to retrieve from the collection
        - **default** - the value to return if the key is not found in the
          collection

        **Returns:**

        A tuple of the `AWSIPv4Prefix`es or `AWSIPv6Prefix`es that contains the
        provided key.
        """
        if isinstance(key, (AWSIPv4Prefix, AWSIPv6Prefix)):
            network = key.prefix
        else:
            network = ip_network(key, strict=False)

        prefixes = list()
        for supernet in supernets(network):
            prefix = self._get_prefix(supernet)
            if prefix:
                prefixes.append(prefix)

        if prefixes:
            prefixes.sort()
            return tuple(prefixes)
        else:
            return default

    def __iter__(self):
        return itertools.chain(self.ipv4_prefixes, self.ipv6_prefixes)

    def __len__(self):
        return len(self.ipv4_prefixes) + len(self.ipv6_prefixes)

    @property
    def regions(self) -> FrozenSet[str]:
        """The set of regions in the collection."""
        if self._regions is None:
            self._regions = frozenset((prefix.region for prefix in self))

        return self._regions

    @property
    def network_border_groups(self) -> FrozenSet[str]:
        """The set of network border groups in the collection."""
        if self._network_border_groups is None:
            self._network_border_groups = frozenset(
                (prefix.network_border_group for prefix in self)
            )

        return self._network_border_groups

    @property
    def services(self) -> FrozenSet[str]:
        """The set of services in the collection.

        The service `"AMAZON"` is not a service but rather an identifier used
        to get all IP address ranges - meaning that every prefix is contained in
        the subset of prefixes tagged with the `"AMAZON"` service. Some IP
        address ranges are only tagged with the `"AMAZON"` service.
        """
        if self._services is None:
            self._services = frozenset(
                (service for prefix in self for service in prefix.services)
            )

        return self._services

    def filter(
        self,
        regions: Union[None, str, Iterable[str]] = None,
        network_border_groups: Union[None, str, Iterable[str]] = None,
        services: Union[None, str, Iterable[str]] = None,
        versions: Union[None, int, Iterable[int]] = None,
    ):
        """Filter the AWS IP address ranges.

        The service `"AMAZON"` is not a service but rather an identifier used
        to get all IP address ranges - meaning that every prefix is contained in
        the subset of prefixes tagged with the `"AMAZON"` service. Some IP
        address ranges are only tagged with the `"AMAZON"` service.

        **Parameters:**

        - **regions** (_optional_ str or iterable sequence of strings) - the
          AWS Regions to include in the subset
        - **network_border_groups** (_optional_ str or iterable sequence of
          strings) - the AWS network border groups to include in the subset
        - **services** (_optional_ str or iterable sequence of strings) - the
          AWS services to include in the subset
        - **versions** (_optional_ int) - the IP address version (4, 6) to
          include in the subset

        **Returns:**

        A new `AWSIPPrefixes` object that contains the subset of IP prefixes that
        match your filter criteria.
        """
        # Normalize, validate, and process the input variables

        # regions
        regions = normalize_to_set(regions) or self.regions
        validate_values("region", regions, valid_values=self.regions)

        # network_border_groups
        network_border_groups = (
            normalize_to_set(network_border_groups) or self.network_border_groups
        )
        validate_values(
            "network_border_group",
            network_border_groups,
            valid_values=self.network_border_groups,
        )

        # services
        services = normalize_to_set(services) or self.services
        validate_values("services", services, valid_values=self.services)

        # prefix_type -> prefix_version
        versions = normalize_to_set(versions) or {4, 6}
        validate_values("versions", versions, valid_values=frozenset((4, 6)))

        # Generate the filtered prefix list
        return self.__class__(
            sync_token=self.sync_token,
            create_date=self.create_date,
            ipv4_prefixes=tuple()
            if 4 not in versions
            else (
                prefix
                for prefix in self.ipv4_prefixes
                if prefix.region in regions
                if prefix.network_border_group in network_border_groups
                if set(prefix.services).intersection(services)
            ),
            ipv6_prefixes=tuple()
            if 6 not in versions
            else (
                prefix
                for prefix in self.ipv6_prefixes
                if prefix.region in regions
                if prefix.network_border_group in network_border_groups
                if set(prefix.services).intersection(services)
            ),
        )



"""Utility functions."""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from ipaddress import (
    IPv4Network,
    IPv6Network,
)
from typing import Any, FrozenSet, Generator, Iterable, Set, Tuple, Type, Union


def check_type(
    variable_name: str,
    obj: Any,
    acceptable_types: Union[Type, Tuple[Type, ...]],
    optional: bool = False,
):
    """Object is an instance of one of the acceptable types or None.
    Args:
        variable_name: The name of the variable being inspected.
        obj: The object to inspect.
        acceptable_types: A type or tuple of acceptable types.
        optional(bool): Whether or not the object may be None.
    Raises:
        TypeError: If the object is not an instance of one of the acceptable
            types, or if the object is None and optional=False.
    """
    assert isinstance(variable_name, str)
    if not isinstance(acceptable_types, tuple):
        acceptable_types = (acceptable_types,)
    assert isinstance(optional, bool)

    if isinstance(obj, acceptable_types):
        # Object is an instance of an acceptable type.
        return
    elif optional and obj is None:
        # Object is None, and that is okay!
        return
    else:
        # Object is something else.
        raise TypeError(
            f"{variable_name} should be a "
            f"{', '.join([t.__name__ for t in acceptable_types])}"
            f"{', or None' if optional else ''}. Received {obj!r} which is a "
            f"{type(obj).__name__}."
        )


def normalize_to_set(
    value: Union[None, str, int, Iterable[Union[str, int]]]
) -> Set[Union[str, int]]:
    """Normalize an optional or iterable variable to a set of unique values."""
    if value is None:
        return set()

    if isinstance(value, (str, int)):
        return {value}

    if isinstance(value, Iterable):
        return set(value)

    raise TypeError("The value must be a string, integer, iterable type, or None.")


def validate_values(
    variable_name: str,
    values: Set[Union[str, int]],
    valid_values: FrozenSet[Union[str, int]],
):
    """Validate the values in a set against a set of valid values."""
    if not values.issubset(valid_values):
        raise ValueError(
            f"One or more of the provided {variable_name} {values!r} do not "
            f"exist in this set of AWS IP address ranges. "
            f"Valid {variable_name}: {valid_values}"
        )


def supernets(
    subnet: Union[IPv4Network, IPv6Network]
) -> Generator[Union[IPv4Network, IPv6Network], None, None]:
    """Incrementally yield the supernets of the provided subnet."""
    for prefix_length in range(subnet.prefixlen, 0, -1):
        yield subnet.supernet(new_prefix=prefix_length)



"""Custom exceptions."""

from typing import Optional, Tuple


class AWSIPRangesException(Exception):
    """Base class for all awsipranges exceptions."""


class HTTPError(AWSIPRangesException):
    """An HTTP/HTTPS error."""

    args: Tuple[object, ...]
    status: Optional[int]
    reason: Optional[str]

    def __init__(self, *args, status: Optional[int], reason: Optional[str]):
        super(HTTPError, self).__init__(*args)
        self.args = args
        self.status = status
        self.reason = reason

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{', '.join([repr(arg) for arg in self.args])}, "
            f"status={self.status!r}, "
            f"reason={self.reason!r}"
            f")"
        )

    def __str__(self):
        msg = []
        if self.status:
            msg.append(str(self.status))

        if self.reason:
            msg.append(self.reason)

        if self.args:
            if msg:
                msg.append("-")
            msg += [str(arg) for arg in self.args]

        return " ".join(msg)


def raise_for_status(response):
    """Raise an HTTPError on 4xx and 5xx status codes."""
    # Get the status code
    if hasattr(response, "status"):
        status = int(response.status)
    elif hasattr(response, "code"):
        status = int(response.code)
    elif hasattr(response, "getstatus"):
        status = int(response.getstatus())
    else:
        raise ValueError(
            f"Response object {response!r} does not contain a status code."
        )

    # Get the URL
    if hasattr(response, "url"):
        url = response.url
    elif hasattr(response, "geturl"):
        url = response.geturl()
    else:
        raise ValueError(f"Response object {response!r} does not contain a url.")

    # Get the reason, if available
    reason = response.reason if hasattr(response, "reason") else None

    if 400 <= status < 500:
        raise HTTPError(f"Client error for URL: {url}", status=status, reason=reason)

    if 500 <= status < 600:
        raise HTTPError(f"Server error for URL: {url}", status=status, reason=reason)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['awsipranges', 'awsipranges.models']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'awsipranges',
    'version': '0.3.1',
    'description': 'Work with the AWS IP address ranges in native Python.',
    'long_description': '# awsipranges\n\n*Work with the AWS IP address ranges in native Python.*\n\n[![License](https://img.shields.io/github/license/aws-samples/awsipranges)](https://github.com/aws-samples/awsipranges/blob/main/LICENSE)\n[![PyPI](https://img.shields.io/pypi/v/awsipranges.svg)](https://pypi.org/project/awsipranges/)\n[![Code Coverage](https://img.shields.io/codecov/c/github/aws-samples/awsipranges)](https://app.codecov.io/github/aws-samples/awsipranges/)\n[![Build](https://img.shields.io/github/workflow/status/aws-samples/awsipranges/tests)](https://github.com/aws-samples/awsipranges/actions/workflows/tests.yml)\n[![Docs](https://img.shields.io/github/workflow/status/aws-samples/awsipranges/publish-docs?label=docs)](https://aws-samples.github.io/awsipranges/)\n\n---\n\nAmazon Web Services (AWS) publishes its [current IP address ranges](https://docs.aws.amazon.com/general/latest/gr/aws-ip-ranges.html) in [JSON](https://ip-ranges.amazonaws.com/ip-ranges.json) format. Python v3 provides an [ipaddress](https://docs.python.org/3/library/ipaddress.html) module in the standard library that allows you to create, manipulate, and perform operations on IPv4 and IPv6 addresses and networks. Wouldn\'t it be nice if you could work with the AWS IP address ranges like native Python objects?\n\n## Features\n\n- Work with the AWS IP prefixes as a simple `AWSIPPrefixes` collection.\n- Quickly check if an IP address, interface, or network is contained in the AWS IP address ranges.\n- Get the AWS IP prefix that contains an IP address, interface, or network.\n- See what services are served from an IP prefix.\n- Filter the AWS IP prefixes by _region_, _network border group_, _service_, and IP prefix _version_.\n- Use the AWS prefix data in your app or automation scripts in the format required by your infrastructure.\n- Easily validate the TLS certificate presented by the IP ranges server.\n- awsipranges has _no third-party dependencies_ and is compatible with CPython v3.7+.\n\n```python\n>>> import awsipranges\n\n>>> aws_ip_ranges = awsipranges.get_ranges(cafile="amazon_root_certificates.pem")\n\n>>> \'52.94.5.15\' in aws_ip_ranges\nTrue\n\n>>> aws_ip_ranges[\'52.94.5.15\']\nAWSIPv4Prefix(\'52.94.5.0/24\', region=\'eu-west-1\', network_border_group=\'eu-west-1\', services=(\'AMAZON\', \'DYNAMODB\'))\n\n>>> aws_ip_ranges.filter(services=\'CODEBUILD\')\n{\'create_date\': datetime.datetime(2021, 8, 24, 1, 31, 14, tzinfo=datetime.timezone.utc),\n \'ipv4_prefixes\': (AWSIPv4Prefix(\'3.26.127.24/29\', region=\'ap-southeast-2\', network_border_group=\'ap-southeast-2\', services=(\'CODEBUILD\',)),\n                   AWSIPv4Prefix(\'3.38.90.8/29\', region=\'ap-northeast-2\', network_border_group=\'ap-northeast-2\', services=(\'CODEBUILD\',)),\n                   AWSIPv4Prefix(\'3.68.251.232/29\', region=\'eu-central-1\', network_border_group=\'eu-central-1\', services=(\'CODEBUILD\',)),\n                   AWSIPv4Prefix(\'3.98.171.224/29\', region=\'ca-central-1\', network_border_group=\'ca-central-1\', services=(\'CODEBUILD\',)),\n                   AWSIPv4Prefix(\'3.101.177.48/29\', region=\'us-west-1\', network_border_group=\'us-west-1\', services=(\'CODEBUILD\',)),\n                   ...),\n \'ipv6_prefixes\': (),\n \'sync_token\': \'1629768674\'}\n\n>>> for prefix in aws_ip_ranges.filter(regions=\'eu-west-1\', services=\'DYNAMODB\'):\n...     print(prefix.network_address, prefix.netmask)\n...\n52.94.5.0 255.255.255.0\n52.94.24.0 255.255.254.0\n52.94.26.0 255.255.254.0\n52.119.240.0 255.255.248.0\n```\n\n## Installation\n\nInstalling and upgrading `awsipranges` is easy:\n\n**Install via PIP**\n\n```shell\n❯ pip install awsipranges\n```\n\n**Upgrade to the latest version**\n\n```shell\n❯ pip install --upgrade awsipranges\n```\n\n## Documentation\n\nExcellent documentation is now available at: https://aws-samples.github.io/awsipranges/\n\nCheck out the [Quickstart](https://aws-samples.github.io/awsipranges/quickstart.html) to dive in and begin using awsipranges.\n\n## Contribute\n\nSee [CONTRIBUTING](https://github.com/aws-samples/awsipranges/blob/main/CONTRIBUTING.md) for information on how to contribute to this project.\n\n## Security\n\nSee [CONTRIBUTING](https://github.com/aws-samples/awsipranges/blob/main/CONTRIBUTING.md#security-issue-notifications) for information on how to report a security issue with this project.\n\n## License\n\nThis project is licensed under the [Apache-2.0 License](https://github.com/aws-samples/awsipranges/blob/main/LICENSE).\n',
    'author': 'Chris Lunsford',
    'author_email': 'cmluns@amazon.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://aws-samples.github.io/awsipranges/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
